"""
This test file tests the lib.config

The lib.config only depends on the database model.
"""
from privacyidea.models import Config, save_config_timestamp, db, NodeName
from .base import MyTestCase
from privacyidea.lib.config import (get_resolver_list,
                                    get_resolver_classes,
                                    get_resolver_class_dict,
                                    get_resolver_types,
                                    get_resolver_module_list,
                                    get_from_config,
                                    get_privacyidea_config,
                                    set_privacyidea_config,
                                    delete_privacyidea_config,
                                    get_token_list,
                                    get_token_module_list,
                                    get_token_class_dict,
                                    get_token_types,
                                    get_token_classes, get_token_prefix,
                                    get_machine_resolver_class_dict,
                                    get_privacyidea_node, get_privacyidea_nodes,
                                    this, get_config_object, invalidate_config_object,
                                    get_multichallenge_enrollable_tokentypes,
                                    get_email_validators,
                                    check_node_uuid_exists)
from privacyidea.lib.resolvers.PasswdIdResolver import IdResolver as PWResolver
from privacyidea.lib.tokens.hotptoken import HotpTokenClass
from privacyidea.lib.tokens.totptoken import TotpTokenClass
import importlib


class ConfigTestCase(MyTestCase):
    """
    Test the config on the database level
    """
    def test_00_get_config(self):
        # set the config
        set_privacyidea_config(key="Hallo", value="What?", typ="string",
                               desc="Some dumb value")

        # get the complete config
        conf = get_from_config()
        self.assertTrue("Hallo" in conf, conf)

        conf = get_from_config("Hallo")
        self.assertTrue(conf == "What?", conf)

        conf = get_from_config("Hello", "Does not exist")
        self.assertTrue(conf == "Does not exist", conf)

        conf = get_privacyidea_config()
        self.assertTrue("Hallo" in conf, conf)

        # delete privacyidea config
        delete_privacyidea_config("Hallo")
        conf = get_from_config("Hallo")
        self.assertFalse(conf == "What?", conf)

        # set more values to create a timestamp and overwrite
        set_privacyidea_config(key="k1", value="v1")
        set_privacyidea_config(key="k2", value="v2")
        set_privacyidea_config(key="k3", value="v3")
        conf = get_from_config("k3")
        self.assertTrue(conf == "v3", conf)
        set_privacyidea_config(key="k3", value="new", typ="string", desc="n")
        conf = get_from_config("k3")
        self.assertTrue(conf == "new", conf)

    def test_01_resolver(self):
        r = get_resolver_list()
        self.assertTrue("privacyidea.lib.resolvers.PasswdIdResolver" in r, r)
        self.assertTrue("privacyidea.lib.resolvers.LDAPIdResolver" in r, r)
        self.assertTrue("privacyidea.lib.resolvers.SCIMIdResolver" in r, r)
        self.assertTrue("privacyidea.lib.resolvers.SQLIdResolver" in r, r)

        # check modules
        mlist = get_resolver_module_list()
        mod_name = "privacyidea.lib.resolvers.PasswdIdResolver"
        module = importlib.import_module(mod_name)

        self.assertTrue(module in mlist, mlist)

        r = get_resolver_classes()
        self.assertTrue(PWResolver in r, r)
        # resolver classes must be available here
        self.assertTrue("pi_resolver_classes" in this.config, this.config)
        self.assertTrue("pi_resolver_types" in this.config, this.config)

        # Class dict
        (classes, types) = get_resolver_class_dict()
        self.assertTrue('privacyidea.lib.resolvers.PasswdIdResolver'
                        '.IdResolver' in classes, classes)
        self.assertTrue(classes.get(
            'privacyidea.lib.resolvers.PasswdIdResolver.IdResolver') ==
                        PWResolver, classes)
        self.assertTrue('privacyidea.lib.resolvers.PasswdIdResolver'
                        '.IdResolver' in types, types)
        self.assertTrue(types.get('privacyidea.lib.resolvers.PasswdIdResolver'
                        '.IdResolver') == "passwdresolver", types)

        # With calling 'get_resolver_classes()' the resolver types will also be cached
        self.assertTrue("pi_resolver_types" in this.config, this.config)
        # When the resolvers are determined, they are stored
        types = get_resolver_types()
        self.assertTrue("passwdresolver" in types, types)

    def test_02_token(self):
        r = get_token_list()
        self.assertTrue("privacyidea.lib.tokens.totptoken" in r, r)
        self.assertTrue("privacyidea.lib.tokens.hotptoken" in r, r)

        # check modules
        mlist = get_token_module_list()
        mod_name = "privacyidea.lib.tokens.totptoken"
        module = importlib.import_module(mod_name)
        self.assertTrue(module in mlist, mlist)

#        r = get_resolver_classes()
#        self.assertTrue(UserResolver in r, r)
#        self.assertTrue(PWResolver in r, r)

        # get_token_class_dict
        (classes, types) = get_token_class_dict()
        self.assertTrue('privacyidea.lib.tokens.hotptoken.HotpTokenClass'
                        in classes, classes)
        self.assertTrue(classes.get(
            'privacyidea.lib.tokens.hotptoken.HotpTokenClass') ==
                        HotpTokenClass, classes)
        self.assertTrue('privacyidea.lib.tokens.totptoken.TotpTokenClass'
                        in classes, classes)
        self.assertTrue(classes.get(
            'privacyidea.lib.tokens.totptoken.TotpTokenClass') ==
                        TotpTokenClass, classes)

        self.assertTrue('privacyidea.lib.tokens.hotptoken.HotpTokenClass'
                        in types, types)
        self.assertTrue('privacyidea.lib.tokens.totptoken.TotpTokenClass'
                        in types, types)
        self.assertTrue(types.get('privacyidea.lib.tokens.hotptoken'
                                  '.HotpTokenClass') == "hotp", types)
        self.assertTrue(types.get('privacyidea.lib.tokens.totptoken'
                                  '.TotpTokenClass') == "totp", types)

        types = get_token_types()
        self.assertTrue("totp" in types, types)
        self.assertTrue("hotp" in types, types)

        # token classes are cached with calling 'get_token_types()'
        self.assertTrue("pi_token_classes" in this.config, this.config)
        self.assertTrue("pi_token_types" in this.config, this.config)
        r = get_token_classes()
        self.assertTrue(TotpTokenClass in r, r)
        self.assertTrue(HotpTokenClass in r, r)

        # Test custom token types
        with self.app_context:
            self.app.config['PI_TOKEN_MODULES'] = 'tests.testdata.fancytoken'
            r = get_token_list()
            self.assertIn("tests.testdata.fancytoken", r, r)
            mlist = get_token_module_list()
            mod = importlib.import_module('tests.testdata.fancytoken')
            self.assertTrue(mod in mlist, mlist)
            (classes, types) = get_token_class_dict()
            self.assertIn('tests.testdata.fancytoken.FancyTokenClass', classes, classes)
            self.assertIn('tests.testdata.fancytoken.FancyTokenClass', types, types)
            self.assertEqual(types['tests.testdata.fancytoken.FancyTokenClass'], 'fancy', types)
            self.app.config.pop('PI_TOKEN_MODULES')

    def test_03_token_prefix(self):
        prefix = get_token_prefix("totp")
        self.assertTrue(prefix == "TOTP", prefix)

        prefix = get_token_prefix("X_Y_Z", "does not exist")
        self.assertTrue(prefix == "does not exist", prefix)

        prefix = get_token_prefix()
        self.assertTrue(prefix.get("totp") == "TOTP", prefix)
        self.assertTrue(prefix.get("hotp") == "OATH", prefix)

    def test_04_store_encrypted_values(self):
        r = set_privacyidea_config("mySecretData", "soho",
                                   typ="password", desc="Very important")
        self.assertTrue(r == "insert", r)

        r = get_from_config("mySecretData")
        self.assertTrue(r == "soho", r)

        r = get_from_config()
        self.assertTrue(r.get("mySecretData") == "soho",
                        r.get("mySecretData"))

    def test_05_machine_resolvers(self):
        (classes, types) = get_machine_resolver_class_dict()
        self.assertTrue("hosts" in types.values(), list(types.values()))
        self.assertTrue("privacyidea.lib.machines.hosts.HostsMachineResolver"
                        in classes, classes)

    def test_06_public_and_admin(self):
        # This tests the new public available config
        set_privacyidea_config("publicInfo1", "info1", typ="public")
        set_privacyidea_config("publicInfo2", "info2", typ="public")
        set_privacyidea_config("secretInfo1", "info1")

        # Get administrators info
        a = get_from_config()
        self.assertTrue("secretInfo1" in a)
        self.assertTrue("publicInfo1" in a)
        a = get_from_config("publicInfo1")
        self.assertEqual(a, "info1")
        a = get_from_config("secretInfo1")
        self.assertEqual(a, "info1")

        # Get public info as user
        a = get_from_config()
        self.assertTrue("publicInfo1" in a)
        a = get_from_config("publicInfo1")
        self.assertEqual(a, "info1")

        # Not able to get private info as user
        a = get_from_config(role="public")
        self.assertTrue("secretInfo1" not in a)
        a = get_from_config("secretInfo1", role="public")
        self.assertEqual(a, None)

    def test_07_node_names(self):
        # if there is no node in the nodename table, the check fails
        self.assertFalse(check_node_uuid_exists("8e4272a9-9037-40df-8aa3-976e4a04b5a9"))

        db.session.add(NodeName(id="8e4272a9-9037-40df-8aa3-976e4a04b5a9", name="Node1"))
        db.session.add(NodeName(id="d1d7fde6-330f-4c12-88f3-58a1752594bf", name="Node2"))
        db.session.commit()

        self.assertTrue(check_node_uuid_exists("8e4272a9-9037-40df-8aa3-976e4a04b5a9"))

        node = get_privacyidea_node()
        self.assertEqual(node, "Node1")
        nodes = get_privacyidea_nodes()
        self.assertIn({"uuid": "8e4272a9-9037-40df-8aa3-976e4a04b5a9", "name": "Node1"}, nodes, nodes)
        self.assertIn({"uuid": "d1d7fde6-330f-4c12-88f3-58a1752594bf", "name": "Node2"}, nodes, nodes)

    def test_08_config_object(self):
        obj1 = get_config_object()
        # Another call to ``get_config_object`` returns the identical config object
        self.assertIs(obj1, get_config_object())
        set_privacyidea_config(key="k1", value="v1")
        # ``set_privacyidea_config`` invalidates the config object, so we get a different one
        obj2 = get_config_object()
        self.assertIsNot(obj1, obj2)
        self.assertEqual(get_config_object().get_config("k1"), "v1")
        # ``set_privacyidea_config`` again invalidates the config object
        set_privacyidea_config(key="k1", value="v2")
        obj3 = get_config_object()
        self.assertIsNot(obj2, obj3)
        self.assertEqual(get_config_object().get_config("k1"), "v2")

    def test_09_invalidate_config_object(self):
        # Test manual invalidation of the config object
        # Ensure we have a config object
        get_config_object()
        # Add a config option *without invalidating the config*
        db.session.add(Config(Key="some_key", Value="some_value"))
        save_config_timestamp(False)
        db.session.commit()
        # The request-local config does not know about the new config option
        self.assertEqual(get_from_config("some_key", "default"), "default")
        # ... so we manually invalidate it ...
        invalidate_config_object()
        # ... and the new config object knows!
        self.assertEqual(get_from_config("some_key", "default"), "some_value")

    def test_10_enrollable_tokentypes(self):
        ttypes = get_multichallenge_enrollable_tokentypes()
        self.assertIn("hotp", ttypes)
        self.assertIn("totp", ttypes)
        self.assertIn("sms", ttypes)
        self.assertIn("email", ttypes)
        self.assertIn("push", ttypes)
        self.assertNotIn("tan", ttypes)
        self.assertNotIn("daplug", ttypes)
        self.assertNotIn("paper", ttypes)

    def test_11_get_email_validators(self):
        ev = get_email_validators()
        self.assertEqual(['tests.testdata.gmailvalidator', 'privacyidea.lib.utils.emailvalidation'], list(ev.keys()))
        validate_email = get_email_validators().get("privacyidea.lib.utils.emailvalidation")
        self.assertTrue(validate_email("valid@email.com"))
        self.assertFalse(validate_email("invalid@email.k"))



#  2023-03-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Init
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
##
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module contains the functions to manage service ids.
It depends on the models.py
"""

import logging

from privacyidea.lib.utils import fetch_one_resource
from privacyidea.lib.error import privacyIDEAError, ResourceNotFoundError
from privacyidea.models import Serviceid, db

log = logging.getLogger(__name__)

ENCODING = "utf-8"


def set_serviceid(name, description=None):
    return Serviceid(name, description).save()


def delete_serviceid(name=None, sid=None):
    """
    Delete the serviceid given by either name or id.
    If there are still applications with this serviceid, the function fails
    with an error.

    :param name: Name of the serviceid to be deleted
    :type name: str
    :param sid: ID of the serviceid to delete
    :type sid: int
    """
    si = None
    if name:
        si = fetch_one_resource(Serviceid, name=name)
    if sid:
        if si:
            if si.id != sid:
                raise privacyIDEAError('ID of the serviceid with name {0!s} does not '
                                       'match given ID ({1:d}).'.format(name, sid))
        else:
            si = fetch_one_resource(Serviceid, id=sid)
    if si:
        # TODO: Implement check for used serviceids
        si.delete()
        db.session.commit()
    else:
        raise ResourceNotFoundError("You need to specify either a ID or name of a serviceid.")


def get_serviceids(name=None, id=None):
    query = Serviceid.query
    if name:
        query = query.filter(Serviceid.name == name)
    if id:
        query = query.filter(Serviceid.id == id)
    return query.all()



import hashlib
from typing import Union

from webauthn import base64url_to_bytes

from privacyidea.lib.token import create_tokenclass_object, log, get_tokens
from privacyidea.lib.tokenclass import ROLLOUTSTATE, TokenClass
from privacyidea.lib.user import User
from privacyidea.models import TokenInfo, Token, Challenge, TokenCredentialIdHash


def get_fido2_token_by_credential_id(credential_id: str) -> Union[TokenClass, None]:
    """
    Find a FIDO2 token (WebAuthn or Passkey) by the credential_id.

    :param credential_id: The credential_id as returned by an authenticator
    :return: The token object or None
    """
    credential_id_hash = hash_credential_id(credential_id)
    try:
        tcih = TokenCredentialIdHash.query.filter(
            TokenCredentialIdHash.credential_id_hash == credential_id_hash).first()
        if tcih:
            db_token = Token.query.filter(Token.id == tcih.token_id).first()
            if db_token:
                return create_tokenclass_object(db_token)
        else:
            log.debug(f"TokenCredentialIdHash entry not found for credential_id {credential_id}. Trying token info...")
            token_id = (TokenInfo.query.filter(TokenInfo.Key == "credential_id_hash")
                        .filter(TokenInfo.Value == credential_id_hash).first().token_id)
            db_token = Token.query.filter(Token.id == token_id).first()
            if db_token:
                # Create a new TokenCredentialIdHash entry for the next time
                tcih = TokenCredentialIdHash(token_id=db_token.id, credential_id_hash=credential_id_hash)
                tcih.save()
                return create_tokenclass_object(db_token)
    except Exception as ex:
        log.warning(f"Error while trying to get token by credential id: {ex}")
    log.warning(f"Failed to find credential with id: {credential_id}.")
    return None


def get_fido2_token_by_transaction_id(transaction_id: str, credential_id: str) -> Union[TokenClass, None]:
    """
    Find a fido2 token (WebAuthn or Passkey) by the transaction_id of the challenge and the credential_id.
    First all challenges are retrieved with the transaction_id. Then the token is searched by the serial of the
    challenge, and lastly, the credential_id of eligible token found is compared with the one provided.
    If the challenge or the token is not found, or the token is not a FIDO2 token, None is returned.

    :param transaction_id: The transaction_id of the challenge
    :param credential_id: The credential_id as returned by an authenticator
    :return: The token object or None
    """
    challenges = Challenge.query.filter(Challenge.transaction_id == transaction_id).all()
    if not challenges:
        log.info(f"No challenges with transaction_id {transaction_id} not found.")
        return None
    token = None
    for challenge in challenges:
        t = Token.query.filter(Token.serial == challenge.serial).first()
        if not t:
            continue
        if t.tokentype == "webauthn":
            possible_token = create_tokenclass_object(t)
            cred_id = possible_token.decrypt_otpkey()
            if cred_id == credential_id:
                token = possible_token
                break
    if not token:
        log.info(f"No fido2 token found for transaction_id {transaction_id}.")
    return token


def get_credential_ids_for_user(user: User) -> list:
    """
    Get a list of credential ids of passkey or webauthn token for a user.
    Can be used to avoid double registration of an authenticator.

    :param user: The user object
    :return: A list of credential ids
    """
    credential_ids = []
    for token in get_tokens(user=user, token_type_list=["passkey"]):
        if token.token.rollout_state != ROLLOUTSTATE.CLIENTWAIT:
            cred_id = token.token.get_otpkey().getKey().decode("utf-8")
            credential_ids.append(cred_id)
    return credential_ids


def hash_credential_id(credential_id: Union[str, bytes]) -> str:
    """
    Hash a credential_id with SHA256 and return the hexdigest.

    :param credential_id: The credential_id to hash
    :return: The hexdigest of the hash
    """
    if isinstance(credential_id, str):
        credential_id = base64url_to_bytes(credential_id)
    return hashlib.sha256(credential_id).hexdigest()


def save_credential_id_hash(credentials_id_hash: str, token_id: int) -> None:
    """
    Save a credential_id hash for a token in the database.

    :param credentials_id_hash: The hash of the credential_id
    :param token_id: The id of the token
    """
    # Check if an entry with that hash already exists
    tcih = TokenCredentialIdHash.query.filter(TokenCredentialIdHash.credential_id_hash == credentials_id_hash).first()
    if tcih:
        token = Token.query.filter(Token.id == tcih.token_id).first()
        if token.id == token_id:
            return
        else:
            # if the token is different, we need to delete the old entry
            log.warning(f"Existing entry in TokenCredentialIdHash for credential_id_hash {credentials_id_hash} and "
                        f"token_id {token.id}. Overwriting it with token_id {token_id}.")
            tcih.delete()
    TokenCredentialIdHash(token_id=token_id, credential_id_hash=credentials_id_hash).save()



"""Add eventhandler tables

Revision ID: 50adc980d625
Revises: 239995464c48
Create Date: 2016-05-04 17:11:51.776440

"""

# revision identifiers, used by Alembic.
revision = '50adc980d625'
down_revision = '239995464c48'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.create_table('eventhandler',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('ordering', sa.Integer(), nullable=False),
        sa.Column('event', sa.Unicode(length=255), nullable=False),
        sa.Column('handlermodule', sa.Unicode(length=255), nullable=False),
        sa.Column('condition', sa.Unicode(length=1024), nullable=True),
        sa.Column('action', sa.Unicode(length=1024), nullable=True),
        sa.PrimaryKeyConstraint('id')
        )
        op.create_table('eventhandleroption',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('eventhandler_id', sa.Integer(), nullable=True),
        sa.Column('Key', sa.Unicode(length=255), nullable=False),
        sa.Column('Value', sa.Unicode(length=2000), nullable=True),
        sa.Column('Type', sa.Unicode(length=2000), nullable=True),
        sa.Column('Description', sa.Unicode(length=2000), nullable=True),
        sa.ForeignKeyConstraint(['eventhandler_id'], ['eventhandler.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('eventhandler_id', 'Key', name='ehoix_1')
        )
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Table 'eventhandler' already exists.")
        else:
            print(exx)
    except Exception as exx:
        print ("Could not add table 'eventhandler'")
        print (exx)


def downgrade():
    op.drop_table('eventhandleroption')
    op.drop_table('eventhandler')




#  product:  privacyIDEA is a fork of LinOTP
#  May, 08 2014 Cornelius Kölbel
#  http://www.privacyidea.org
#
#  2014-12-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             remove remnant code and code cleanup during
#             flask migration. Ensure code coverage.
#  2014-10-19 Remove class SecurityModule from __init__.py
#             and add it here.
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  product:  LinOTP2
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contains the crypto functions as implemented by the default security module,
which is the encryption key in the file.

The contents of the file is tested in tests/test_lib_crypto.py
"""

import logging
import binascii
import os

from hashlib import sha256

from privacyidea.lib.crypto import (geturandom, zerome, aes_cbc_encrypt,
                                    aes_cbc_decrypt)
from privacyidea.lib.error import HSMException
from privacyidea.lib.utils import (is_true, to_unicode, to_bytes,
                                   hexlify_and_unicode)

from .password import PASSWORD
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives import padding

log = logging.getLogger(__name__)


def create_key_from_password(password):
    """
    Create a key from the given password.
    This is used to encrypt and decrypt the enckey file.

    :param password:
    :type password: str or bytes
    :return: the generated key
    :rtype: bytes
    """
    key = sha256(to_bytes(password)).digest()[0:32]
    return key


def int_list_to_bytestring(int_list):  # pragma: no cover
    return b"".join([bytes((i, )) for i in int_list])


class SecurityModule(object):
    TOKEN_KEY = 0
    CONFIG_KEY = 1
    VALUE_KEY = 2
    DEFAULT_KEY = 3

    mapping = {
        'token': TOKEN_KEY,
        'config': CONFIG_KEY,
        'value': VALUE_KEY
    }

    is_ready = False

    def __init__(self, config=None):
        log.error("This is the base class. You should implement this!")
        self.config = config
        self.name = "SecurityModule"

    def setup_module(self, params):
        fname = 'setup_module'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented {0!s}".format(fname))

    ''' base methods '''
    def random(self, length):
        fname = 'random'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented {0!s}".format(fname))

    def encrypt(self, data, iv, key_id=TOKEN_KEY):
        fname = 'encrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented {0!s}".format(fname))

    def decrypt(self, enc_data, iv, key_id=TOKEN_KEY):
        fname = 'decrypt'
        log.error("This is the base class. You should implement "
                  "the method : %s " % (fname,))
        raise NotImplementedError("Should have been implemented {0!s}".format(fname))

    def decrypt_password(self, crypt_pass):
        """
        Decrypt the given password. The CONFIG_KEY is used to decrypt it.

        :param crypt_pass: the encrypted password with the leading iv,
            separated by the ':', hexlified
        :type crypt_pass: str

        :return: decrypted data
        :rtype: str
        """
        return self._decrypt_value(crypt_pass, self.CONFIG_KEY)

    def decrypt_pin(self, crypt_pin):
        """
        Decrypt the given encrypted PIN with the TOKEN_KEY

        :param crypt_pin: the encrypted pin with the leading iv,
            separated by the ':'
        :type crypt_pin: str

        :return: decrypted data
        :rtype: str
        """
        return self._decrypt_value(crypt_pin, self.TOKEN_KEY)

    def encrypt_password(self, password):
        """
        Encrypt the given password with the CONFIG_KEY and a random IV.

        This function returns a unicode string with a
        hexlified contents of the IV and the encrypted data separated by a
        colon like "4956:44415441"

        :param password: The password that is to be encrypted
        :type password: str
        :return: encrypted data - leading iv, separated by the ':'
        :rtype: str
        """
        return self._encrypt_value(password, self.CONFIG_KEY)

    def encrypt_pin(self, pin):
        """
        Encrypt the given PIN with the TOKEN_KEY and a random IV

        This function returns a unicode string with a
        hexlified contents of the IV and the encrypted data separated by a
        colon like "4956:44415441"

        :param pin: the pin that should be encrypted
        :type pin: str
        :return: encrypted data - leading iv, separated by the ':'
        :rtype: str
        """
        return self._encrypt_value(pin, self.TOKEN_KEY)

    ''' base methods for pin and password '''
    def _encrypt_value(self, value, slot_id):
        """
        base method to encrypt a value
        - uses one slot id to encrypt a string
        returns as string with leading iv, separated by ':'

        :param value: the value that is to be encrypted
        :type value: str

        :param slot_id: slot of the key array
        :type slot_id: int

        :return: encrypted data with leading iv and separator ':'
        :rtype: str
        """
        iv = self.random(16)
        v = self.encrypt(to_bytes(value), iv, slot_id)

        cipher_value = binascii.hexlify(iv) + b':' + binascii.hexlify(v)
        return cipher_value.decode("utf-8")

    def _decrypt_value(self, crypt_value, slot_id):
        """
        base method to decrypt a value
        - used one slot id to encrypt a string with leading iv, separated by ':'

        :param crypt_value: the the value that is to be decrypted
        :type crypt_value: str

        :param slot_id: slot of the key array
        :type slot_id: int

        :return: decrypted data
        :rtype: str
        """
        # split at ":"
        pos = crypt_value.find(':')
        bIV = crypt_value[:pos]
        bData = crypt_value[pos + 1:]

        iv = binascii.unhexlify(bIV)
        data = binascii.unhexlify(bData)

        clear_data = self.decrypt(data, iv, slot_id)

        return to_unicode(clear_data)

    def create_keys(self):
        """
        This can be used to create the encryption keys
        :return: Module dependent
        """
        fname = "create_keys"
        raise NotImplementedError("Should have been implemented {0!s}".format(fname))


class DefaultSecurityModule(SecurityModule):

    def __init__(self, config=None):
        """
        Init of the default security module. The config needs to contain the key
        file. The key file can be encrypted, then the config also needs to
        provide the information, that the key file is encrypted.

           {"file": "/etc/secretkey",
            "crypted": True}

        If the key file is encrypted, the HSM is not immediately ready. It will
        return HSM.is_ready == False.
        Then the function "setup_module({"password": "PW to decrypt"}) needs
        to be called.

        :param config: contains the configuration definition
        :type  config: dict

        :return -
        """
        config = config or {}
        self.name = "Default"
        self.config = config
        self.crypted = False
        self.is_ready = True
        self._id = binascii.hexlify(os.urandom(3))

        if "file" not in config:
            log.error("No secret file defined. A parameter "
                      "PI_ENCFILE is missing in your pi.cfg.")
            raise HSMException("no secret file defined: PI_ENCFILE!")

        # We determine, if the file is encrypted.
        with open(config.get("file"), 'rb') as f:
            cipher = f.read()

        if len(cipher) > 100:
            config["crypted"] = True

        if is_true(config.get("crypted", "")):
            self.crypted = True
            self.is_ready = False

        self.secFile = config.get('file')
        self.secrets = {}

    def _get_secret(self, slot_id=SecurityModule.TOKEN_KEY, password=None):
        """
        internal function, which reads the key from the defined
        slot in the file. It also caches the encryption key to the dictionary
        self.secrets.

        If the file is encrypted, the encryption key is decrypted with the
        password and also cached in self.secrets.

        :param slot_id: slot id of the key array
        :type slot_id: int

        :return: key or secret
        :rtype: binary string
        """
        slot_id = int(slot_id)
        if self.crypted and slot_id in self.secrets:
            return self.secrets.get(slot_id)

        secret = b''

        if self.crypted:
            # if the password was not provided, read it from the module
            # singleton cache
            password = password or PASSWORD
            if not password:
                raise HSMException("Error decrypting the encryption key. "
                                   "No password provided!")
            # Read all keys, decrypt them and return the key for
            # the slot id
            # TODO we assume here, that the file contains the hexlified data
            with open(self.secFile) as f:
                cipher = f.read()

            try:
                keys = self.password_decrypt(cipher, password)
            except UnicodeDecodeError as e:
                raise HSMException("Error decrypting the encryption key. You "
                                   "probably provided the wrong password.")
            secret = keys[slot_id*32:(slot_id+1)*32]

        else:
            # Only read the key with the slot_id
            with open(self.secFile, 'rb') as f:
                for _i in range(0, slot_id + 1):
                    secret = f.read(32)

            if secret == b"":
                raise HSMException("No secret key defined for index: %s !\n"
                                   "Please extend your %s"" !"
                                   % (str(slot_id), self.secFile))

        # cache the result
        self.secrets[slot_id] = secret
        return secret

    def setup_module(self, params):
        """
        callback, which is called during the runtime to initialze the
        security module.

        E.g. here the password for an encrypted keyfile can be provided like::

           {"password": "top secreT"}

        :param params: The password for the key file
        :type  params: dict

        :return: -
        """
        if self.crypted is False:
            return
        if "password" in params:
            PASSWORD = params.get("password")
        else:
            raise HSMException("missing password")

        # if we have a crypted file and a password, we take all keys
        # from the file and put them in a hash
        # After this we do not require the password anymore
        for handle in [self.TOKEN_KEY, self.CONFIG_KEY, self.VALUE_KEY]:
            # fill self.secrets
            self.secrets[handle] = self._get_secret(handle, PASSWORD)

        self.is_ready = True
        return self.is_ready

    # the real interfaces: random, encrypt, decrypt
    @staticmethod
    def random(length=32):
        """
        Create and return random bytes.

        :param length: length of the random byte array
        :type length: int

        :return: random bytes
        :rtype: byte string
        """
        return os.urandom(length)

    def encrypt(self, data, iv, key_id=SecurityModule.TOKEN_KEY):
        """
        security module methods: encrypt

        :param data: the data that is to be encrypted
        :type data: bytes

        :param iv: initialisation vector
        :type iv: bytes

        :param key_id: slot of the key array. The key file contains 96
            bytes, which are made up of 3 32byte keys.
        :type key_id: int

        :return: encrypted data
        :rtype:  bytes
        """
        if self.is_ready is False:
            raise HSMException('setup of security module incomplete')

        key = self._get_secret(key_id)

        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(data) + padder.finalize()

        res = aes_cbc_encrypt(key, iv, padded_data)

        if self.crypted is False:
            zerome(key)
            del key
        return res

    @staticmethod
    def password_encrypt(data, password):
        """
        Encrypt the given text with the password.
        A key is derived from the password and used to encrypt the text in
        AES MODE_CBC. The IV is returned together with the cipher text.
        <IV:Cipher>

        :param data: The text to encrypt
        :type data: str or bytes
        :param password: The password to derive a key from
        :type password: str or bytes
        :return: IV and cipher text
        :rtype: str
        """
        bkey = create_key_from_password(password)
        # convert input to ascii, so we can securely append bin data for padding
        input_data = binascii.hexlify(to_bytes(data))
        input_data += b"\x01\x02"
        padding = (16 - len(input_data) % 16) % 16
        input_data += padding * b"\0"
        iv = geturandom(16)
        cipher = aes_cbc_encrypt(bkey, iv, input_data)
        iv_hex = hexlify_and_unicode(iv)
        cipher_hex = hexlify_and_unicode(cipher)
        return "{0!s}:{1!s}".format(iv_hex, cipher_hex)

    @staticmethod
    def password_decrypt(enc_data, password):
        """
        Decrypt the given data with the password.
        A key is derived from the password. The data is hexlified data, the IV
        is the first part, separated with a ":".

        :param enc_data: The hexlified data
        :type enc_data: str
        :param password: The password, that is used to decrypt the data
        :type password: str or bytes
        :return: The clear test
        :rtype: bytes
        """
        bkey = create_key_from_password(password)
        # split the input data
        iv_hex, cipher_hex = enc_data.strip().split(":")
        iv_bin = binascii.unhexlify(iv_hex)
        cipher_bin = binascii.unhexlify(cipher_hex)
        output = aes_cbc_decrypt(bkey, iv_bin, cipher_bin)
        # remove padding
        eof = output.rfind(b"\x01\x02")
        if eof >= 0:
            output = output[:eof]
        cleartext = binascii.unhexlify(output)
        return cleartext

    def decrypt(self, enc_data, iv, key_id=SecurityModule.TOKEN_KEY):
        """
        Decrypt the given data with the key from the key slot

        :param enc_data: the to be decrypted data
        :type enc_data: bytes

        :param iv: initialisation vector (salt)
        :type iv: bytes

        :param key_id: slot of the key array
        :type key_id: int

        :return: decrypted data
        :rtype: bytes
        """
        if self.is_ready is False:
            raise HSMException('setup of security module incomplete')

        key = self._get_secret(key_id)
        output = aes_cbc_decrypt(key, iv, enc_data)

        try:
            unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
            data = unpadder.update(output) + unpadder.finalize()
        except ValueError as _e:
            # try legacy padding
            # remove padding
            eof = output.rfind(b"\x01\x02")
            if eof >= 0:
                output = output[:eof]

            # convert output from ascii, back to bin data
            data = binascii.unhexlify(output)

        if self.crypted is False:
            zerome(key)
            del key

        return data



#  2018-08-01 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             EventCounter Task
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import logging
from privacyidea.lib.task.base import BaseTask
from privacyidea.lib.monitoringstats import write_stats
from privacyidea.lib.counter import reset, read
from privacyidea.lib.utils import is_true
from privacyidea.lib import _


__doc__ = """This task module reads event counters and writes them to the 
MonitoringStats database table."""

log = logging.getLogger(__name__)


class EventCounterTask(BaseTask):
    identifier = "EventCounter"
    description = "Write snapshots of the event counter to the MonitoringStats table."

    @property
    def options(self):
        return {
            "event_counter": {
                "type": "str",
                "description": _("The name of the event counter to read."),
                "required": True
            },
            "stats_key": {
                "type": "str",
                "description": _("The name of the stats key to write to the MonitoringStats "
                                 "table."),
                "required": True
            },
            "reset_event_counter": {
                "type": "bool",
                "description": _("Whether to reset the event_counter, if it is read and written "
                                 "to the MonitoringStats table.")
            }

        }

    def do(self, params):
        event_counter = params.get("event_counter")
        stats_key = params.get("stats_key")
        reset_event_counter = params.get("reset_event_counter")

        counter_value = read(event_counter)
        if is_true(reset_event_counter) and counter_value:
            reset(event_counter)

        # now write the current value of the counter
        if counter_value is None:
            log.warning("Trying to create statistics of a counter_value '{0}', "
                        "that does not exist.".format(event_counter))
        else:
            write_stats(stats_key, counter_value)

        return True




"""
This test case test the REST API
api/applications.py
"""

import json
from .base import MyApiTestCase


class APIApplicationsResolverTestCase(MyApiTestCase):

    def test_get_applications(self):
        with self.app.test_request_context('/application/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            value = result.get("value")
            self.assertIn("ssh", value)
            self.assertIn("luks", value)
            sshtokenoptions = value.get("ssh").get("options")
            lukstokenoptions = value.get("luks").get("options")
            offlinetokenoptions = value.get("offline").get("options")
            # Check the correct token types
            self.assertIn("sshkey", sshtokenoptions)
            self.assertIn("totp", lukstokenoptions)
            self.assertIn("hotp", offlinetokenoptions)
            self.assertIn("webauthn", offlinetokenoptions)
            self.assertIn("service_id", sshtokenoptions.get("sshkey"))
            self.assertIn("user", sshtokenoptions.get("sshkey"))
            self.assertEqual("str", sshtokenoptions.get("sshkey").get("service_id").get("type"))
            self.assertEqual("str", sshtokenoptions.get("sshkey").get("user").get("type"))



"""
This tests the files
  lib/task/simplestats.py
"""
from privacyidea.lib.user import User
from privacyidea.lib.tokenclass import TOKENKIND
from privacyidea.lib.token import init_token
from privacyidea.models import db
from .base import MyTestCase
from privacyidea.lib.monitoringstats import get_values
from flask import current_app

from privacyidea.lib.task.simplestats import SimpleStatsTask

simple_results = {'total_tokens': (1, 2, 3, 4),
                  'hardware_tokens': (0, 1, 2, 2),
                  'software_tokens': (1, 1, 1, 2),
                  'assigned_tokens': (0, 0, 1, 1),
                  'unassigned_hardware_tokens': (0, 1, 1, 1),
                  'user_with_token': (0, 0, 1, 2)
                  }


class TaskSimpleStatsTestCase(MyTestCase):

    serials = ['SE1', 'SE2', 'SE3', 'SE4']
    otpkey = "3132333435363738393031323334353637383930"

    def test_00_read_simplestats_to_monitoringstats(self):
        self.setUp_user_realms()

        # create a simple statistics class
        sst = SimpleStatsTask(current_app.config)
        # and set all parameters to 'true'
        params = {}
        for o in sst.options.keys():
            params[o] = True

        # first we create a software token
        init_token({"type": "totp", "otpkey": self.otpkey, "serial": self.serials[0]})

        sst.do(params)
        db.session.commit()
        for o in sst.options.keys():
            self.assertEqual(simple_results[o][0], get_values(o)[0][1],
                             msg="Current option: {0}".format(o))

        # add a hardware token
        init_token({"type": "totp", "otpkey": self.otpkey, "serial": self.serials[1]},
                   tokenkind=TOKENKIND.HARDWARE)

        sst.do(params)
        db.session.commit()
        for o in sst.options.keys():
            self.assertEqual(simple_results[o][1], get_values(o)[1][1],
                             msg="Current option: {0}".format(o))

        # add a hardware token and assign it to a user
        token = init_token({"type": "totp", "otpkey": self.otpkey, "serial": self.serials[2]},
                           tokenkind=TOKENKIND.HARDWARE,
                           user=User(login="cornelius", realm=self.realm1))
        self.assertEqual("cornelius", token.user.login)
        self.assertTrue(token.is_active())
        self.assertEqual(TOKENKIND.HARDWARE, token.get_tokeninfo('tokenkind'))

        sst.do(params)
        db.session.commit()
        for o in sst.options.keys():
            self.assertEqual(simple_results[o][2], get_values(o)[2][1],
                             msg="Current option: {0}".format(o))

        # add a software token and assign it to a user
        token = init_token({"type": "totp", "otpkey": self.otpkey, "serial": self.serials[3]},
                           user=User(login="selfservice", realm=self.realm1))
        self.assertEqual("selfservice", token.user.login)
        self.assertTrue(token.is_active())
        self.assertEqual(TOKENKIND.SOFTWARE, token.get_tokeninfo('tokenkind'))

        # check if getting only certain stats works
        params['assigned_tokens'] = False
        sst.do(params)
        db.session.commit()
        self.assertEqual(3, len(get_values('assigned_tokens')))
        self.assertEqual(4, len(get_values('user_with_token')))
        for o in sst.options.keys():
            if o != 'assigned_tokens':
                self.assertEqual(simple_results[o][3], get_values(o)[3][1],
                                 msg="Current option: {0}".format(o))
        self.assertEqual(simple_results['assigned_tokens'][3],
                         get_values('assigned_tokens')[2][1])




"""
This test file tests the lib.tokens.smstoken
"""

from .base import MyTestCase, FakeFlaskG, FakeAudit
from privacyidea.lib.resolver import (save_resolver)
from privacyidea.lib.realm import (set_realm)
from privacyidea.lib.user import (User)
from privacyidea.lib.tokens.indexedsecrettoken import IndexedSecretTokenClass, PIIXACTION
from privacyidea.lib.policy import set_policy, delete_policy, SCOPE, ACTION, PolicyClass
from privacyidea.models import Token
from privacyidea.lib.token import init_token, remove_token

PWFILE = "tests/testdata/passwords"


class IndexedSecretTokenTestCase(MyTestCase):
    """
    Test the IndexedSecret Token
    """
    email = "pi_tester@privacyidea.org"
    otppin = "topsecret"
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    resolvername3 = "reso3"
    realm1 = "realm1"
    realm2 = "realm2"
    serial1 = "SE123456"
    serial2 = "SE000000"
    otpkey = "3132333435363738393031323334353637383930"

    success_body = "ID 12345"

    def test_00_create_user_realm(self):
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1, [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm1,
                    resolver=self.resolvername1)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.resolver1@realm1>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm1', resolver='resolver1')"
        self.assertTrue(user_repr == expected, user_repr)

    def test_01_create_token(self):
        my_secret = "mySecretInformation"
        db_token = Token(self.serial1, tokentype="indexedsecret")
        db_token.save()
        token = IndexedSecretTokenClass(db_token)
        token.update({"otpkey": my_secret})
        token.save()
        serial = token.get_serial()
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "indexedsecret", token.token)
        self.assertTrue(token.type == "indexedsecret", token.type)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "PIIX", class_prefix)
        self.assertTrue(token.get_class_type() == "indexedsecret", token)

        # Create a challenge
        r, message, transaction_id, reply_dict = token.create_challenge()
        attribute = reply_dict.get("attributes")
        self.assertTrue(r)
        self.assertIn("Please enter the position", message)

        password_list = [my_secret[x - 1] for x in attribute.get("random_positions")]
        password = "".join(password_list)
        # Wrong transaction_id
        r = token.check_challenge_response(passw=password, options={"transaction_id": "wrong"})
        self.assertEqual(-1, r)

        # wrong password - wrong length
        r = token.check_challenge_response(passw="wrong", options={"transaction_id": transaction_id})
        self.assertEqual(-1, r)

        # wrong password - wrong contents
        r = token.check_challenge_response(passw="XX", options={"transaction_id": transaction_id})
        self.assertEqual(-1, r)

        # Successful authentication, we can also pass the transaction_id in the state.
        r = token.check_challenge_response(passw=password, options={"state": transaction_id})
        self.assertEqual(1, r)

        db_token.delete()

    def test_02_init_token(self):
        # Create the tokenclass via init_token
        my_secret = "mysimplesecret"
        t = init_token({"type": "indexedsecret",
                        "otpkey": my_secret,
                        "serial": "PIIX1234"})
        self.assertEqual(t.token.tokentype, "indexedsecret")
        self.assertEqual(t.token.serial, "PIIX1234")

        remove_token("PIIX1234")

    def test_03_challenge_text_position_count(self):
        # test challenge text and position count
        my_secret = "mysimplesecret"
        set_policy("pol1", scope=SCOPE.AUTH, action="indexedsecret_{0!s}=5".format(PIIXACTION.COUNT))
        set_policy("pol2", scope=SCOPE.AUTH,
                   action="indexedsecret_challenge_text=Hier sind die Positionen: {0!s}")

        t = init_token({"type": "indexedsecret",
                        "otpkey": my_secret,
                        "serial": "PIIX1234"})
        g = FakeFlaskG()
        g.audit_object = FakeAudit
        g.policy_object = PolicyClass()

        # Create a challenge
        r, message, transaction_id, reply_dict = t.create_challenge(options={"g": g})
        attribute = reply_dict.get("attributes")
        # The challenge text from the policy is used.
        self.assertIn("Hier sind die Positionen:", message)
        password_list = [my_secret[x - 1] for x in attribute.get("random_positions")]
        password = "".join(password_list)
        # The password has length 5, due to the pol2
        self.assertEqual(5, len(password))
        # Successful auth
        r = t.check_challenge_response(passw=password, options={"transaction_id": transaction_id})
        self.assertEqual(1, r)

        delete_policy("pol1")
        delete_policy("pol2")
        remove_token("PIIX1234")



"""v3.11: Add table tokencredentialidhash

Revision ID: 903a6ed6f6c4
Revises: c128c01a5520
Create Date: 2025-02-19 11:53:45.910016

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError

# revision identifiers, used by Alembic.
revision = '903a6ed6f6c4'
down_revision = 'c128c01a5520'


def upgrade():
    try:
        op.create_table('tokencredentialidhash',
                        sa.Column('id', sa.Integer(), primary_key=True),
                        sa.Column('credential_id_hash', sa.String(length=256), nullable=False),
                        sa.Column('token_id', sa.Integer(), nullable=False),
                        sa.ForeignKeyConstraint(['token_id'], ['token.id'], ),
                        sa.Index('ix_tokencredentialidhash_credentialidhash',
                                 'credential_id_hash', unique=True))
    except (OperationalError, ProgrammingError) as ex:
        if "already exists" in str(ex.orig).lower():
            print("Table 'tokencredentialidhash' already exists.")
        else:
            print("Could not add table 'tokencredentialidhash' to database.")
            raise


def downgrade():
    try:
        op.drop_table('tokencredentialidhash')
    except (OperationalError, ProgrammingError) as exx:
        msg = str(exx.orig).lower()
        if "no such table" in msg or "unknown table" in msg or "does not exist" in msg:
            print("Table 'tokencredentialidhash' already removed.")
        else:
            print("Could not remove table 'tokencredentialidhash'.")
            raise



"""v3.6: add fraction to MySQL DateTime type in audit table

Revision ID: 3ba618f6b820
Revises: 59ef3e03bc62
Create Date: 2021-05-05 11:27:13.705851

"""

# revision identifiers, used by Alembic.
revision = '3ba618f6b820'
down_revision = '59ef3e03bc62'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql


def upgrade():
    try:
        if op._proxy.migration_context.dialect.name == 'mysql':
            with op.batch_alter_table('pidea_audit') as batch_op:
                batch_op.alter_column('date', type_=mysql.DATETIME(fsp=6),
                                      existing_type=mysql.DATETIME)
                batch_op.alter_column('startdate', type_=mysql.DATETIME(fsp=6),
                                      existing_type=mysql.DATETIME)
                batch_op.alter_column('duration', type_=mysql.DATETIME(fsp=6),
                                      existing_type=mysql.DATETIME)
    except Exception as exx:
        print("Could not add fraction to MySQL DateTime Type in audit table.")
        print(exx)


def downgrade():
    try:
        if op._proxy.migration_context.dialect.name == 'mysql':
            with op.batch_alter_table('pidea_audit') as batch_op:
                batch_op.alter_column('date', type_=mysql.DATETIME,
                                      existing_type=mysql.DATETIME(fsp=6))
                batch_op.alter_column('startdate', type_=mysql.DATETIME,
                                      existing_type=mysql.DATETIME(fsp=6))
                batch_op.alter_column('duration', type_=mysql.DATETIME,
                                      existing_type=mysql.DATETIME(fsp=6))
    except Exception as exx:
        print("Could not remove fraction from MySQL DateTime Type in audit table.")
        print(exx)



"""
This file contains the tests for the info module lib/info.py
"""

from .base import MyTestCase
from privacyidea.lib.info.rss import get_news, RSS_FEEDS


class RSSTestCase(MyTestCase):

    def test_01_check_definition(self):
        self.assertIn("Community News", RSS_FEEDS)
        self.assertIn("privacyIDEA News", RSS_FEEDS)
        self.assertIn("NetKnights News", RSS_FEEDS)

    def test_01_fetch_news(self):
        r = get_news()
        # Check, that we can fetch the three basic news feeds
        self.assertIn("Community News", r)
        self.assertIn("privacyIDEA News", r)
        self.assertIn("NetKnights News", r)

    def test_02_arbitrary_news(self):
        r = get_news({"Heise": "https://www.heise.de/newsticker/heise.rdf"})
        self.assertIn("Heise", r)

    def test_03_fail_fetching(self):
        r = get_news({"Broken": "https://netknights.it"})
        self.assertIn("Broken", r)
        # Since it is broken, it has an empty list.
        self.assertEqual(r.get("Broken"), [])



"""Add priority to policy

Revision ID: 204d8d4f351e
Revises: 145ce80decd
Create Date: 2018-05-09 15:30:39.878219

"""

# revision identifiers, used by Alembic.
revision = '204d8d4f351e'
down_revision = '145ce80decd'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        # As we defined 'priority' to be non-NULLable, we supply a server_default value of 1
        # here to set the priority of all existing policies to 1.
        op.add_column('policy', sa.Column('priority', sa.Integer(), nullable=False, server_default='1'))
    except Exception as exx:
        print("Could not add column 'priority' to table 'policy'.")
        print(exx)

def downgrade():
    op.drop_column('policy', 'priority')



"""
Based on the database models, which are tested in tests/test_db_model.py,
there are different modules.

resolver.py contains functions to simply deal with resolver definitions.
On this level users and realms are not know, yet.

realm.py contains functions to deal with realm. Realms are a list of several
resolvers. So prior to bother the realm.py, the resolver.py should be
understood and working.
On this level, users are not known, yet.

user.py contains functions to deal with users. A user object is an entity
in a realm. And of course the user object itself can be found in a resolver.
But you need to have working resolver.py and realm.py to be able to
work with user.py
"""

from privacyidea.lib.framework import _
from flask_babel import lazy_gettext
__all__ = ['_', 'lazy_gettext']



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
import json

from privacyidea.lib.error import ParameterError
from privacyidea.models import TokenContainer


class ContainerTemplateBase:
    template_option_values = {}

    def __init__(self, db_template):
        self._db_template = db_template

    @classmethod
    def get_class_type(cls) -> str:
        """
        Returns the type of the container template class.
        """
        return "generic"

    @classmethod
    def get_supported_token_types(cls) -> list:
        """
        Returns the supported token types for this container template.
        """
        supported_token_types = ["hotp", "remote", "daypassword", "spass", "totp", "4eyes", "paper", "push",
                                 "indexedsecret", "tan", "applspec", "registration", "sms", "email", "tiqr"]
        supported_token_types.sort()
        return supported_token_types

    @property
    def name(self) -> str:
        return self._db_template.name

    @property
    def container_type(self) -> str:
        return self._db_template.container_type

    @classmethod
    def get_template_class_options(cls) -> dict:
        return cls.template_option_values

    @classmethod
    def get_template_option_keys(cls) -> list:
        return cls.template_option_values.keys()

    @property
    def template_options(self) -> str:
        return self._db_template.options

    def get_template_options_as_dict(self) -> dict:
        return json.loads(self.template_options) if self.template_options else {}

    @template_options.setter
    def template_options(self, options: dict):
        if not isinstance(options, dict):
            raise ParameterError("options must be a dict")
        validated_options = {}

        # Validates token types in options
        supported_token_types = self.get_supported_token_types()
        tokens = options.get("tokens", [])
        for token in tokens:
            token_type = token.get("type", None)
            if token.get("type", None) not in supported_token_types:
                raise ParameterError(f"Unsupported token type {token_type} for {self.get_class_type()} templates!")
        validated_options["tokens"] = tokens

        self._db_template.options = json.dumps(validated_options)
        self._db_template.save()

    @property
    def id(self) -> int:
        return self._db_template.id

    @property
    def default(self) -> bool:
        return self._db_template.default

    @default.setter
    def default(self, value: bool):
        if not isinstance(value, bool):
            raise ParameterError("Default must be a boolean")
        self._db_template.default = value
        self._db_template.save()

    @property
    def containers(self) -> list:
        return self._db_template.containers

    def delete(self):
        self._db_template.delete()



import logging
from testfixtures import LogCapture

from .base import MyApiTestCase
from privacyidea.lib.policy import set_policy, SCOPE, ACTION, delete_policy, CONDITION_SECTION
from privacyidea.lib.token import remove_token
from privacyidea.models import db, NodeName


class APIPolicyTestCase(MyApiTestCase):
    def test_00_get_policy(self):
        with self.app.test_request_context('/policy/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            self.assertEqual(res.json["result"]["value"], [], res.json)

        # test the policy export
        # first without policies (this used to fail due to an index error)
        with self.app.test_request_context('/policy/export/pols.txt',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual(res.mimetype, 'text/plain', res)
            self.assertEqual(res.content_length, 0, res)

        # test export with a given policy
        set_policy("hide_welcome", scope=SCOPE.WEBUI, action=ACTION.HIDE_WELCOME, client="10.1.2.3")
        with self.app.test_request_context('/policy/export/pols.txt',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual(res.mimetype, 'text/plain', res)
            self.assertGreater(res.content_length, 0, res)
            self.assertIn(b'[hide_welcome]', res.data, res)
        delete_policy('hide_welcome')

    def test_01_set_policy(self):
        with self.app.test_request_context('/policy/pol1',
                                           method='POST',
                                           data={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "check_all_resolvers": "true",
                                                 "realm": "realm1",
                                                 "priority": 3,
                                                 "description": "This is a test policy"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertTrue("setPolicy pol1" in result.get("value"),
                            result.get("value"))

        # get the policies and see if check_all_resolvers and priority are set
        with self.app.test_request_context('/policy/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            pol1 = value[0]
            self.assertEqual(pol1.get("check_all_resolvers"), True)
            self.assertEqual(pol1.get("priority"), 3)
            self.assertEqual(pol1.get("description"), "This is a test policy", pol1)

        # get active policies
        with self.app.test_request_context('/policy/?active=true',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            self.assertEqual(len(value), 1)
            pol1 = value[0]
            self.assertEqual(pol1.get("check_all_resolvers"), True)
            self.assertEqual(pol1.get("priority"), 3)

        # Update policy to check_all_resolvers = false and priority = 5
        with self.app.test_request_context('/policy/pol1',
                                           method='POST',
                                           data={
                                               "action": ACTION.NODETAILFAIL,
                                               "client": "10.1.2.3",
                                               "scope": SCOPE.AUTHZ,
                                               "check_all_resolvers": "false",
                                               "priority": 5,
                                               "realm": "realm1"},
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertTrue("setPolicy pol1" in result.get("value"),
                            result.get("value"))

        # get the policies and see if check_all_resolvers and priority are set
        with self.app.test_request_context('/policy/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            pol1 = value[0]
            self.assertEqual(pol1.get("check_all_resolvers"), False)
            self.assertEqual(pol1.get("priority"), 5)

        delete_policy("pol1")

        with self.app.test_request_context('/policy/polpinode',
                                           method='POST',
                                           data={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "pinode": "Node1",
                                                 "priority": 1,
                                                 "realm": "realm1"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertTrue("setPolicy polpinode" in result.get("value"),
                            result.get("value"))

        # get the policies and see if the pinode was set
        with self.app.test_request_context('/policy/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            pol1 = value[0]
            self.assertEqual(pol1.get("pinode"), ["Node1"])

        delete_policy("polpinode")

    def test_02_set_policy_conditions(self):
        # set a policy with conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "realm": "realm1",
                                                 "conditions": [
                                                     ["userinfo", "groups", "contains", "group1", True],
                                                     ["userinfo", "type", "equals", "secure", False],
                                                     ["HTTP header", "Origin", "equals", "https://localhost", True]
                                                 ]},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # get the policy and check its conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            cond1 = value[0]
            self.assertEqual(cond1["realm"], ["realm1"])
            self.assertEqual(len(cond1["conditions"]), 3)
            # order of conditions is not guaranteed
            self.assertIn(["userinfo", "groups", "contains", "group1", True], cond1["conditions"])
            self.assertIn(["userinfo", "type", "equals", "secure", False], cond1["conditions"])
            self.assertIn(["HTTP header", "Origin", "equals", "https://localhost", True], cond1["conditions"])

        # update the policy, but not its conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # get the policy and check its conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            cond1 = value[0]
            self.assertEqual(cond1["realm"], ["realm2"])
            self.assertEqual(len(cond1["conditions"]), 3)
            # order of conditions is not guaranteed
            self.assertIn(["userinfo", "groups", "contains", "group1", True], cond1["conditions"])
            self.assertIn(["userinfo", "type", "equals", "secure", False], cond1["conditions"])
            self.assertIn(["HTTP header", "Origin", "equals", "https://localhost", True], cond1["conditions"])

        # update the policy conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [["userinfo", "type", "equals", "secure", True]],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # get the policy and check its conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            cond1 = value[0]
            self.assertEqual(cond1["realm"], ["realm2"])
            self.assertEqual(len(cond1["conditions"]), 1)
            # order of conditions is not guaranteed
            self.assertIn(["userinfo", "type", "equals", "secure", True], cond1["conditions"])

        # test some invalid conditions
        # no 5-tuples
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [["userinfo", "type", "equals", "secure"]],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400)

        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [["userinfo", "type", "equals",
                                                                 "secure", True, "extra"]],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400)

        # wrong types
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [["userinfo", "type", "equals", 123, False]],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400)

        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [["userinfo", "type", "equals", "123", "true"]],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400)

        # reset conditions
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILFAIL,
                                                 "client": "10.1.2.3",
                                                 "scope": SCOPE.AUTHZ,
                                                 "conditions": [],
                                                 "realm": "realm2"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        with self.app.test_request_context('/policy/cond1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            self.assertTrue(res.json['result']['status'], res.json)
            value = res.json['result']['value']
            cond1 = value[0]
            self.assertEqual(cond1["realm"], ["realm2"])
            self.assertEqual(cond1["conditions"], [])

        # delete policy
        with self.app.test_request_context('/policy/cond1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

    def test_03_set_adminuser_policy(self):
        # Set a policy for the user testadmin
        with self.app.test_request_context('/policy/pol1adminuser',
                                           method='POST',
                                           data={
                                               "action": "{0!s}, {1!s}".format(ACTION.POLICYDELETE,
                                                                               ACTION.POLICYREAD),
                                               "scope": SCOPE.ADMIN,
                                               "realm": "",
                                               "adminuser": "testadmin",
                                               "client": "10.1.2.3"},
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            self.assertTrue("setPolicy pol1adminuser" in result.get("value"),
                            result.get("value"))

        # Get the policies
        with self.app.test_request_context('/policy/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            policies = result.get("value")
            self.assertEqual(1, len(policies))
            self.assertEqual("pol1adminuser", policies[0].get("name"))

        # The admin is not allowed to enroll a token!
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           data={"genkey": 1},
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            # Forbidden
            self.assertTrue(res.status_code == 403, res)
            data = res.json
            result = data.get("result")
            status = result.get("status")
            self.assertFalse(status)
            message = result.get("error").get("message")
            self.assertEqual(message, "Admin actions are defined, but you are not allowed to enroll this token type!")

        # Delete the policy
        with self.app.test_request_context('/policy/pol1adminuser',
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            status = result.get("status")
            self.assertTrue(status)

    def test_04_policy_defs(self):
        db.session.add(NodeName(id="8e4272a9-9037-40df-8aa3-976e4a04b5a9", name="Node1"))
        db.session.add(NodeName(id="d1d7fde6-330f-4c12-88f3-58a1752594bf", name="Node2"))
        db.session.commit()
        with self.app.test_request_context('/policy/defs/conditions',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            status = result.get("status")
            self.assertTrue(status)
            value = result.get("value")
            self.assertIn("comparators", value)
            self.assertIn("sections", value)

        with self.app.test_request_context('/policy/defs/pinodes',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            result = data.get("result")
            status = result.get("status")
            self.assertTrue(status)
            value = result.get("value")
            self.assertIn("Node1", value)
            self.assertIn("Node2", value)

    def test_05_invalid_client(self):
        with self.app.test_request_context('/policy/condFalse',
                                           method='POST',
                                           json={
                                               "client": "10.1.2.3.4",
                                               "action": ACTION.NODETAILSUCCESS,
                                               "realm": "realm1",
                                               "conditions": [[CONDITION_SECTION.HTTP_REQUEST_HEADER,
                                                               "User-Agent", "broken",
                                                               "SpecialApp", True]],
                                               "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 400)
            data = res.json
            result = data.get("result")
            self.assertEqual(result['error'], {'code': 302, 'message': 'ERR302: Invalid client definition!'})


class APIPolicyConditionTestCase(MyApiTestCase):

    def test_01_check_httpheader_condition(self):
        self.setUp_user_realms()
        # enroll a simple pass token
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           json={"type": "spass", "pin": "1234",
                                                 "serial": "sp1", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # test an auth request
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertTrue("detail" in result)
            self.assertEqual(result.get("detail").get("message"), "matching 1 tokens")

        # set a policy with conditions
        # Request from a certain user agent will not see the detail
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILSUCCESS,
                                                 "client": "10.1.2.3",
                                                 "realm": "realm1",
                                                 "conditions": [[CONDITION_SECTION.HTTP_REQUEST_HEADER,
                                                                 "User-Agent", "equals", "SpecialApp", True]],
                                                 "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # A request with another header will display the details
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           headers={"User-Agent": "somethingelse"},
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertTrue("detail" in result)
            self.assertEqual(result.get("detail").get("message"), "matching 1 tokens")

        # A request with the dedicated header will not display the details
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           headers={'User-Agent': 'SpecialApp'},
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertFalse("detail" in result)

        # A request without such a header
        with LogCapture(level=logging.ERROR) as lc:
            with self.app.test_request_context('/validate/check',
                                               method='POST',
                                               headers={"Another": "header"},
                                               json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                     "client": "10.1.2.3"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, 403)
                result = res.json
                self.assertIn("Unknown HTTP header key referenced in condition of policy 'cond1'",
                              result["result"]["error"]["message"])
                # Make sure the missing key is described in the error log
                lc.check_present(("privacyidea.lib.policy", "ERROR",
                                  "Unknown HTTP header key 'User-Agent' referenced in condition of policy 'cond1'"))

        # A request without such a specific header - always has a header
        with LogCapture(level=logging.ERROR) as lc:
            with self.app.test_request_context('/validate/check',
                                               method='POST',
                                               json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                     "client": "10.1.2.3"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, 403)
                result = res.json
                self.assertIn("Unknown HTTP header key referenced in condition of policy 'cond1'",
                              result["result"]["error"]["message"])
                # Make sure the missing key is described in the error log
                lc.check_present(("privacyidea.lib.policy", "ERROR",
                                  "Unknown HTTP header key 'User-Agent' referenced in condition of policy 'cond1'"))

        # Test http header policy with broken matching
        # update the policy
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILSUCCESS,
                                                 "client": "10.1.2.3",
                                                 "realm": "realm1",
                                                 "conditions": [[CONDITION_SECTION.HTTP_REQUEST_HEADER,
                                                                 "User-Agent", "broken",
                                                                 "SpecialApp", True]],
                                                 "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
        # now test the policy
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           headers={"User-Agent": "SpecialApp"},
                                           json={"pass": "1234", "user": "cornelius",
                                                 "realm": "realm1", "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = res.json
            self.assertIn("Invalid comparison in the HTTP header conditions of policy",
                          result["result"]["error"]["message"])

        # Also check for an unknown section
        # update the policy
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILSUCCESS,
                                                 "client": "10.1.2.3",
                                                 "realm": "realm1",
                                                 "conditions": [['blabla',
                                                                 "User-Agent", "equals",
                                                                 "SpecialApp", True]],
                                                 "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
        # now test the policy
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           headers={"User-Agent": "SpecialApp"},
                                           json={"pass": "1234", "user": "cornelius",
                                                 "realm": "realm1", "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403)
            result = res.json
            # The text can be "Policy u'cond1' has condition with unknown section"
            # or "Policy 'cond1' has condition with unknown section"
            # so we only match for a substring
            self.assertIn("has condition with unknown section",
                          result["result"]["error"]["message"], result)

        delete_policy("cond1")
        remove_token("sp1")

    def test_02_check_httpenvironment_condition(self):
        self.setUp_user_realms()
        # enroll a simple pass token
        with self.app.test_request_context('/token/init',
                                           method='POST',
                                           json={"type": "spass", "pin": "1234",
                                                 "serial": "sp1", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # test an auth request
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertTrue("detail" in result)
            self.assertEqual(result.get("detail").get("message"), "matching 1 tokens")

        # set a policy with conditions
        # Request with a certain request method will not see the user details
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILSUCCESS,
                                                 "realm": "realm1",
                                                 "client": "10.1.2.3",
                                                 "conditions": [[CONDITION_SECTION.HTTP_ENVIRONMENT,
                                                                 "REQUEST_METHOD", "equals", "POST", True]],
                                                 "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        # A GET request will contain the details!
        with self.app.test_request_context('/validate/check',
                                           method='GET',
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertTrue("detail" in result)
            self.assertEqual(result.get("detail").get("message"), "matching 1 tokens")

        # A POST request will NOT contain the details!
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                 "client": "10.1.2.3"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)
            result = res.json
            self.assertFalse("detail" in result)

        delete_policy("cond1")
        # Now we run a test with a non-existing environment key
        with self.app.test_request_context('/policy/cond1',
                                           method='POST',
                                           json={"action": ACTION.NODETAILSUCCESS,
                                                 "realm": "realm1",
                                                 "client": "10.1.2.3",
                                                 "conditions": [[CONDITION_SECTION.HTTP_ENVIRONMENT,
                                                                 "NON_EXISTING", "equals", "POST", True]],
                                                 "scope": SCOPE.AUTHZ},
                                           headers={'PI-Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200)

        with LogCapture(level=logging.ERROR) as lc:
            with self.app.test_request_context('/validate/check',
                                               method='POST',
                                               json={"pass": "1234", "user": "cornelius", "realm": "realm1",
                                                     "client": "10.1.2.3"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(res.status_code, 403)
                result = res.json
                self.assertIn("Unknown HTTP environment key referenced in condition of policy 'cond1'",
                              result["result"]["error"]["message"])
                # Make sure the missing key is described in the error log
                lc.check_present(("privacyidea.lib.policy", "ERROR",
                                  "Unknown HTTP environment key 'NON_EXISTING' referenced "
                                  "in condition of policy 'cond1'"))

        delete_policy("cond1")
        remove_token("sp1")



"""
This file contains the tests for the job queue modules.

In particular, this tests
lib/queue/*.py
"""
from huey import RedisHuey
import mock

from privacyidea.app import create_app
from privacyidea.config import TestingConfig
from privacyidea.lib.error import ServerError
from privacyidea.lib.queue import job, JOB_COLLECTOR, JobCollector, get_job_queue, wrap_job, has_job_queue
from privacyidea.lib.queues.huey_queue import HueyQueue
from privacyidea.lib.queues.base import QueueError
from .base import OverrideConfigTestCase, MyTestCase


class TestSender(object):
    """ defined in order to be able to mock the ``send_mail`` function in tests """
    def send_mail(*args, **kwargs):
        pass


SENDER = TestSender()


@job("test.my_add")
def my_add(a, b):
    return a + b


@job("test.my_send_mail")
def my_send_mail(message):
    SENDER.send_mail(message)
    return 1337


class NoQueueTestCase(OverrideConfigTestCase):
    class Config(TestingConfig):
        PI_JOB_QUEUE_CLASS = ""

    def test_01_no_job_queue(self):
        self.assertFalse(has_job_queue())
        with self.assertRaises(ServerError):
            get_job_queue()

    def test_02_collector(self):
        self.assertIsInstance(JOB_COLLECTOR, JobCollector)
        test_dict = {"test.my_add": (my_add, (), {}),
                     "test.my_send_mail": (my_send_mail, (), {})}
        try:
            self.assertTrue(test_dict.viewitems() <= JOB_COLLECTOR.jobs.viewitems())
        except AttributeError:
            self.assertTrue(test_dict.items() <= JOB_COLLECTOR.jobs.items())


class InvalidQueueTestCase(MyTestCase):
    def test_01_create_app_graciously(self):
        class Config(TestingConfig):
            PI_JOB_QUEUE_CLASS = "obviously.invalid"

        with mock.patch.dict("privacyidea.config.config", {"testing": Config}):
            app = create_app("testing", "") # we do not throw an exception


class HueyQueueTestCase(OverrideConfigTestCase):
    class Config(TestingConfig):
        PI_JOB_QUEUE_CLASS = "privacyidea.lib.queues.huey_queue.HueyQueue"
        PI_JOB_QUEUE_NAME = "myqueuename"
        PI_JOB_QUEUE_IMMEDIATE = True  # avoid redis server for testing

    def test_01_app_job_queue(self):
        queue = get_job_queue()
        self.assertIsInstance(queue, HueyQueue)
        self.assertEqual(queue.options, {"name": "myqueuename", "immediate": True})
        self.assertTrue({"test.my_add", "test.my_send_mail"}.issubset(set(queue.jobs)))
        self.assertIsInstance(queue.huey, RedisHuey)
        self.assertEqual(queue.huey.name, "myqueuename")
        self.assertFalse(queue.huey.store_none)
        with self.assertRaises(QueueError):
            queue.register_job("test.my_add", lambda x: x)

    def test_03_enqueue_jobs(self):
        queue = get_job_queue()
        queue.enqueue("test.my_add", (3, 4), {}) # No result is stored

        with mock.patch.object(SENDER, 'send_mail') as mock_mail:
            queue.enqueue("test.my_send_mail", ("hi",), {})
            mock_mail.assert_called_once_with("hi")

        with self.assertRaises(QueueError):
            queue.enqueue("test.unknown", ("hi",), {})

    def test_04_wrap_jobs(self):
        wrapped = wrap_job("test.my_send_mail", True)
        with mock.patch.object(SENDER, 'send_mail') as mock_mail:
            result = wrapped("hi")
            mock_mail.assert_called_once_with("hi")
            self.assertTrue(result)
        with mock.patch.object(SENDER, 'send_mail') as mock_mail:
            result = my_send_mail("hi")
            mock_mail.assert_called_once_with("hi")
            self.assertEqual(result, 1337)


"""
This test file tests the lib/machine.py for attaching and detaching tokens
"""
from privacyidea.lib.error import ResourceNotFoundError, ParameterError

HOSTSFILE = "tests/testdata/hosts"
from .base import MyTestCase
from privacyidea.lib.machine import (attach_token, detach_token, add_option,
                                     delete_option, list_machine_tokens,
                                     list_token_machines, get_auth_items)
from privacyidea.lib.token import init_token, get_tokens
from privacyidea.lib.machineresolver import save_resolver


sshkey = "ssh-rsa " \
         "AAAAB3NzaC1yc2EAAAADAQABAAACAQDJy0rLoxqc8SsY8DVAFijMsQyCv" \
         "hBu4K40hdZOacXK4O6OgnacnSKN56MP6pzz2+4svzvDzwvkFsvf34pbsgD" \
         "F67PPSCsimmjEQjf0UfamBKh0cl181CbPYsph3UTBOCgHh3FFDXBduPK4DQz" \
         "EVQpmqe80h+lsvQ81qPYagbRW6fpd0uWn9H7a/qiLQZsiKLL07HGB+NwWue4os" \
         "0r9s4qxeG76K6QM7nZKyC0KRAz7CjAf+0X7YzCOu2pzyxVdj/T+KArFcMmq8V" \
         "dz24mhcFFXTzU3wveas1A9rwamYWB+Spuohh/OrK3wDsrryStKQv7yofgnPMs" \
         "TdaL7XxyQVPCmh2jVl5ro9BPIjTXsre9EUxZYFVr3EIECRDNWy3xEnUHk7Rzs" \
         "734Rp6XxGSzcSLSju8/MBzUVe35iXfXDRcqTcoA0700pIb1ANYrPUO8Up05v4" \
         "EjIyBeU61b4ilJ3PNcEVld6FHwP3Z7F068ef4DXEC/d7pibrp4Up61WYQIXV/" \
         "utDt3NDg/Zf3iqoYcJNM/zIZx2j1kQQwqtnbGqxJMrL6LtClmeWteR4420uZx" \
         "afLE9AtAL4nnMPuubC87L0wJ88un9teza/N02KJMHy01Yz3iJKt3Ou9eV6kqO" \
         "ei3kvLs5dXmriTHp6g9whtnN6/Liv9SzZPJTs8YfThi34Wccrw== " \
         "NetKnights GmbH"


class MachineTokenTestCase(MyTestCase):
    """
    Test the attaching of tokens to machines
    """
    serial = "myToken"
    resolvername = "reso1"
    serial2 = "ser1"
    serialHotp = "hotp2"

    def test_00_setup(self):
        token1 = init_token({"type": "spass", "serial": self.serial})
        resolver1 = save_resolver({"name": self.resolvername,
                                   "type": "hosts",
                                   "filename": HOSTSFILE})
        init_token({"type": "hotp", "serial": self.serialHotp, "otpkey": "313233"})

    def test_01_attach_token(self):
        mt = attach_token(self.serial, "luks", hostname="gandalf")
        self.assertEqual(mt.token.serial, self.serial)
        self.assertEqual(mt.token.machine_list[0].machine_id, "192.168.0.1")

        # look at token, if we see the machine.
        tok = get_tokens(serial=self.serial)[0]
        self.assertEqual(tok.token.machine_list[0].machine_id, "192.168.0.1")

        # problem attaching token with incomplete machine definition (missing
        #  resolver)
        self.assertRaises(Exception, attach_token, self.serial, "luks",
                          machine_id="192.168.0.1")

        # attach token with non-existing serial
        self.assertRaises(ResourceNotFoundError, attach_token, "nonexisting", "luks")

        # attach token with non-existing application
        test_token = init_token({"type": "spass"})
        self.assertRaises(ParameterError, attach_token, test_token.get_serial(), "nonexisting")
        # look at token, if we do not see the machine
        tok = get_tokens(serial=test_token.get_serial())[0]
        machine_list = tok.token.machine_list
        self.assertEqual(len(machine_list), 0)

    def test_02_detach_token(self):
        detach_token(self.serial, "luks", hostname="gandalf")

        # look at token, if we do not see the machine
        tok = get_tokens(serial=self.serial)[0]
        machine_list = tok.token.machine_list
        self.assertEqual(len(machine_list), 0)

    def test_03_add_delete_option(self):
        mt = attach_token(self.serial, "luks", hostname="gandalf")
        self.assertEqual(mt.token.serial, self.serial)
        self.assertEqual(mt.token.machine_list[0].machine_id, "192.168.0.1")

        r = add_option(serial=self.serial, application="luks",
                       hostname="gandalf", options={"option1": "value1",
                                                    "option2": "valü2"})
        self.assertEqual(r, 2)

        # The options are accessible via the Token!!!
        tok = get_tokens(serial=self.serial)[0]
        option_list = tok.token.machine_list[0].option_list
        self.assertEqual(len(option_list), 2)
        for option in option_list:
            if option.mt_key == "option1":
                self.assertEqual(option.mt_value, "value1")
            elif option.mt_key == "option2":
                self.assertEqual(option.mt_value, "valü2")
            else:
                self.fail("Unspecified Option! {0!s}".format(option.mt_key))

        r = delete_option(serial=self.serial, application="luks",
                          hostname="gandalf", key="option1")
        self.assertEqual(r, 1)

        # The options are accessible via the Token!!!
        tok = get_tokens(serial=self.serial)[0]
        option_list = tok.token.machine_list[0].option_list
        self.assertEqual(len(option_list), 1)

    def test_04_list_tokens_for_machine(self):
        serial = "serial2"
        init_token({"type": "spass", "serial": serial})
        mt = attach_token(serial, "luks", hostname="gandalf")

        tokenlist = list_machine_tokens(hostname="gandalf")

        self.assertEqual(len(tokenlist), 2)

    def test_05_list_machines_for_token(self):
        machinelist = list_token_machines(self.serial)

        self.assertEqual(len(machinelist), 1)
        self.assertEqual(machinelist[0].get("resolver"), "reso1")
        self.assertEqual(machinelist[0].get("machine_id"), "192.168.0.1")

    def test_10_auth_items(self):
        # create an SSH token
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": sshkey})
        self.assertEqual(token_obj.type, "sshkey")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(hostname="gandalf", serial=self.serial2,
                         application="ssh", options={"user": "testuser"})

        self.assertEqual(r.machine_id, "192.168.0.1")

        # fetch the auth_items for application SSH on machine gandalf
        ai = get_auth_items("gandalf", application="ssh")
        sshkey_auth_items = ai.get("ssh")
        self.assertEqual(len(sshkey_auth_items), 1)
        self.assertTrue(sshkey_auth_items[0].get("sshkey").startswith(
            "ssh-rsa"))

        # fetch the auth_items with user restriction for SSH
        ai = get_auth_items("gandalf", application="ssh",
                            filter_param={"user": "testuser"})
        sshkey_auth_items = ai.get("ssh")
        self.assertEqual(len(sshkey_auth_items), 1)
        self.assertTrue(sshkey_auth_items[0].get("sshkey").startswith(
            "ssh-rsa"))

        # try to fetch SSH keys for user, who has no ssh keys
        ai = get_auth_items("gandalf", application="ssh",
                            filter_param={"user": "nonExist"})
        sshkey_auth_items = ai.get("ssh")
        # None or an empty list
        self.assertFalse(sshkey_auth_items)
        # detach token
        detach_token(self.serial2, "ssh", hostname="gandalf")
        mt = list_token_machines(self.serial2)
        self.assertEqual(0, len(mt))

    def test_11_attach_token_without_machine(self):
        mt = attach_token(self.serialHotp, "offline")
        self.assertEqual(mt.token.serial, self.serialHotp)

        # look at token, if we see the machine.
        tok = get_tokens(serial=self.serialHotp)[0]
        self.assertEqual(tok.token.machine_list[0].machine_id, None)
        self.assertEqual(tok.token.machine_list[0].application, "offline")

    def test_12_list_tokens_without_machine(self):
        tokenlist = list_machine_tokens(serial=self.serialHotp, application="offline")
        self.assertEqual(len(tokenlist), 1)

    def test_13_get_authitems_without_machine(self):
        # fetch the auth_items for application offline and token serialHotp
        ai = get_auth_items(application="offline", serial=self.serialHotp)
        offline_auth_items = ai.get("offline")
        self.assertEqual(len(offline_auth_items), 1)
        offline_auth_item = offline_auth_items[0]
        self.assertIn("refilltoken", offline_auth_item)
        self.assertIn("response", offline_auth_item)
        # The counter of the HOTP token is set to >100
        tok = get_tokens(serial=self.serialHotp)[0]
        self.assertEqual(tok.token.count, 100)

    def test_14_detach_token_without_machine(self):
        detach_token(self.serialHotp, "offline")

        # look at token, if we do not see the machine
        tok = get_tokens(serial=self.serialHotp)[0]
        machine_list = tok.token.machine_list
        self.assertEqual(len(machine_list), 0)

    def test_15_detach_ssh_by_service_id(self):
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": sshkey})
        self.assertEqual(token_obj.type, "sshkey")

        # Attach the token to the machine "gandalf" with the application SSH
        r = attach_token(serial=self.serial2,
                         application="ssh", options={"user": "testuser", "service_id": "webserver"})
        r = attach_token(serial=self.serial2,
                         application="ssh", options={"user": "root", "service_id": "mailserver"})
        r = attach_token(serial=self.serial2,
                         application="ssh", options={"user": "testuser", "service_id": "mailserver"})
        mt = list_token_machines(self.serial2)
        self.assertEqual(3, len(mt), mt)

        # Detach only one application
        detach_token(self.serial2, "ssh", filter_params={"user": "testuser", "service_id": "mailserver"})
        mt = list_token_machines(self.serial2)
        self.assertEqual(2, len(mt))

        # Detach all remaining applications
        detach_token(self.serial2, "ssh")
        mt = list_token_machines(self.serial2)
        self.assertEqual(0, len(mt))



"""Base test configuration to set up/teardown tests."""
import unittest
import mock
from sqlalchemy.orm.session import close_all_sessions

from privacyidea.app import create_app
from privacyidea.config import TestingConfig
from privacyidea.models import db, save_config_timestamp
from privacyidea.lib.resolver import save_resolver
from privacyidea.lib.realm import set_realm
from privacyidea.lib.user import User
from privacyidea.lib.auth import create_db_admin
from privacyidea.lib.auditmodules.base import Audit
from privacyidea.lib.lifecycle import call_finalizers


PWFILE = "tests/testdata/passwords"
PWFILE2 = "tests/testdata/passwd"


class FakeFlaskG(object):
    policy_object = None
    logged_in_user = {}
    audit_object = None
    client_ip = None
    request_headers = None
    serial = None

    def get(self, name, default=None):
        return self.__dict__.get(name, default)


class FakeAudit(Audit):

    def __init__(self):
        super(FakeAudit).__init__()
        self.audit_data = {}


class MyTestCase(unittest.TestCase):
    app = None
    app_context = None
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    resolvername3 = "reso3"
    realm1 = "realm1"
    realm2 = "realm2"
    realm3 = "realm3"
    realm4 = "realm4"
    testadmin = 'testadmin'
    testadminpw = 'testpw'
    testadminmail = "admin@test.tld"
    serials = ["SE1", "SE2", "SE3"]
    otpkey = "3132333435363738393031323334353637383930"
    valid_otp_values = ["755224",
                        "287082",
                        "359152",
                        "969429",
                        "338314",
                        "254676",
                        "287922",
                        "162583",
                        "399871",
                        "520489"]

    @classmethod
    def setUpClass(cls):
        cls.app = create_app('testing', "")
        cls.app_context = cls.app.app_context()
        cls.app_context.push()
        db.create_all()

        # save the current timestamp to the database to avoid hanging cached data
        save_config_timestamp()
        db.session.commit()
        # Create an admin for tests.
        create_db_admin(cls.testadmin, cls.testadminmail, cls.testadminpw)

    def tearDown(self):
        # Rollback uncommitted changes to the DB and close the session to
        # avoid breaking following tests due to unfinished transactions
        try:
            db.session.commit()
        finally:
            db.session.rollback()
        db.session.close()

    def setUp_user_realms(self):
        # create user realm
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1, [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm1,
                    resolver=self.resolvername1)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.resolver1@realm1>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm1', resolver='resolver1')"
        self.assertTrue(user_repr == expected, user_repr)

    def setUp_user_realm2(self):
        # create user realm
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm2,
                                    [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm2,
                    resolver=self.resolvername1)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.resolver1@realm2>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm2', resolver='resolver1')"
        self.assertTrue(user_repr == expected, user_repr)

    def setUp_user_realm3(self):
        # create user realm
        rid = save_resolver({"resolver": self.resolvername3,
                             "type": "passwdresolver",
                             "fileName": PWFILE2})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm3,
                                    [{'name': self.resolvername3}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        user = User(login="root",
                    realm=self.realm3,
                    resolver=self.resolvername3)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.reso3@realm3>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm3', resolver='reso3')"
        self.assertTrue(user_repr == expected, user_repr)

    def setUp_user_realm4_with_2_resolvers(self):
        # create user realm
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)
        rid = save_resolver({"resolver": self.resolvername3,
                             "type": "passwdresolver",
                             "fileName": PWFILE2})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm4,
                                    [{'name': self.resolvername1}, {'name': self.resolvername3}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 2)

        user = User(login="root",
                    realm=self.realm4,
                    resolver=self.resolvername3)

        user_str = "{0!s}".format(user)
        self.assertTrue(user_str == "<root.reso3@realm4>", user_str)

        self.assertFalse(user.is_empty())
        self.assertTrue(User().is_empty())

        user_repr = "{0!r}".format(user)
        expected = "User(login='root', realm='realm4', resolver='reso3')"
        self.assertTrue(user_repr == expected, user_repr)

    def setUp_sqlite_resolver_realm(self, sqlite_file, realm):
        parameters = {'resolver': "sqlite_resolver",
                      "type": "sqlresolver",
                      'Driver': 'sqlite',
                      'Server': '/tests/testdata/',
                      'Database': sqlite_file,
                      'Table': 'users',
                      'Encoding': 'utf8',
                      'Editable': True,
                      'Map': """{ "username": "username",
                        "userid" : "id",
                        "email" : "email",
                        "surname" : "name",
                        "givenname" : "givenname",
                        "password" : "password",
                        "phone": "phone",
                        "mobile": "mobile"}"""
                      }
        r = save_resolver(parameters)
        self.assertTrue(r)
        success, fail = set_realm(realm, [{'name': "sqlite_resolver"}])
        self.assertEqual(len(success), 1)
        self.assertEqual(len(fail), 0)

    @classmethod
    def tearDownClass(cls):
        call_finalizers()
        close_all_sessions()
        db.drop_all()
        db.engine.dispose()
        cls.app_context.pop()

    def authenticate(self):
        with self.app.test_request_context('/auth',
                                           data={"username": self.testadmin,
                                                 "password": self.testadminpw},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            self.at = result.get("value").get("token")

    def authenticate_selfservice_user(self):
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), res.data)
            # In self.at_user we store the user token
            self.at_user = result.get("value").get("token")
            # check that this is a user
            role = result.get("value").get("role")
            self.assertTrue(role == "user", result)
            self.assertEqual(result.get("value").get("realm"), "realm1")

    def find_most_recent_audit_entry(self, **filter_data):
        """
        Given audit log entry filters, return the most recent entry matching the criteria,
        or raise an IndexError if there is no such entry.
        This is useful for testing the audit log behavior.
        """
        sorted_filter = filter_data.copy()
        sorted_filter["sortorder"] = "desc"
        with self.app.test_request_context('/audit/',
                                           method='GET',
                                           data=sorted_filter,
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res.data)
            self.assertTrue(res.is_json, res)
            result = res.json['result']
            self.assertIn('auditdata', result['value'])
            # return the last entry
            return res.json["result"]["value"]["auditdata"][0]


class OverrideConfigTestCase(MyTestCase):
    """
    helper class that allows to modify the app config processed by ``create_app``.
    This can be useful if config values need to be adjusted *for app creation*.
    For that, just override the inner ``Config`` class.
    """
    class Config(TestingConfig):
        pass

    @classmethod
    def setUpClass(cls):
        """ override privacyidea.config.config["testing"] with the inner config class """
        with mock.patch.dict("privacyidea.config.config", {"testing": cls.Config}):
            MyTestCase.setUpClass()


class MyApiTestCase(MyTestCase):
    @classmethod
    def cls_auth(cls, app):
        with app.test_request_context('/auth', data={"username": cls.testadmin,
                                                     "password": cls.testadminpw},
                                      method='POST'):
            res = app.full_dispatch_request()
            assert res.status_code == 200
            result = res.json.get("result")
            assert result.get("status")
            cls.at = result.get("value").get("token")

    @classmethod
    def setUpClass(cls):
        super(MyApiTestCase, cls).setUpClass()
        cls.cls_auth(cls.app)
